#from otr_staff import TabPage
import sys
from os import path
from otr_glyph import TabGlyphList
from german_page import *
from german_glyph import *
import plugins
from gamera.core import *
from gamera.args import *
from gamera.gui import gui
from gamera import toolkit
from gamera import knn
from gamera.toolkits.otr.otr_staff import *
from gamera.toolkits.otr.otr_glyph import symbol_table
from gamera.toolkits.otr.plugins import *

import re
from gamera.core import *
from gamera.gui import has_gui

if has_gui.has_gui:
    from gamera.gui import var_name
    import wx
    import otr_module_icon

    class GermanLutePageModuleIcon(toolkit.CustomIcon):

        def __init__(self, *args, **kwargs):
            toolkit.CustomIcon.__init__(self, *args, **kwargs)
            self.classes = ["GermanPage"]

        def get_icon():
            return toolkit.CustomIcon.to_icon(\
                    otr_module_icon.getBitmap())
        get_icon = staticmethod(get_icon)

        
        def check(data):
            import inspect
            return inspect.ismodule(data) and\
                    data.__name__.endswith("otr")
        check = staticmethod(check)

        def right_click(self, parent, event, shell):
            pass
            self._shell=shell
            x, y=event.GetPoint()
            menu=wx.Menu()

            # create the menu entry for each class listed in
            # 'classes' (they all point to the same method but
            # can be distinguished by their menu index)
            index = wx.NewId()
            menu.Append(index, "Train French/Italian Tab Page")
            wx.EVT_MENU(parent, index, self.openfrenchtrainingsdialog)
            index = wx.NewId()
            menu.Append(index, "Train German Tab Page")
            wx.EVT_MENU(parent, index, self.opengermantrainingsdialog)
            index = wx.NewId()
            menu.Append(index, "Create GermanPage object")
            wx.EVT_MENU(parent, index, self.createGermanLutePageObj)
            parent.PopupMenu(menu, wx.Point(x, y))

        def double_click(self):
            pass

   




#--------------------------------------------------------------------------

        def opengermantrainingsdialog(self,event):
            class c_opt:
                imagefile = ""
                datafile = ""
                bar_detection=True
                sym_tab = ""
                specklesize = 0
                smooth = 0
                removelyrics = None
                corr_rot = None
                remove_cpy = None
                fill=False
            opt = c_opt()


            # start options dialog
            dialog=Args([FileOpen("Image file", opt.imagefile, "*.*"),
                         FileOpen("Training data file for reuse", opt.datafile, "*.xml"),
                         FileOpen("Symbol_table", opt.sym_tab, "*.txt"),
                         Check("Heuristic bar detection",default=True),
                         Check("Fill holes in lines"),
                         Int("Despeckling size", default=3),
                         Choice("Smoothing method", ["median", "closing", "wspeckle", "none"]),
                         Check("Correct rotation"),
                         Check("Remove copy border")],
                        name="Select training options")
            params=dialog.show()
            if not params:
                return

            # wxPython
            import wx

            # map parameters
            i=0
            opt.imagefile=params[i]; i+=1
            opt.datafile=params[i]; i+=1
            opt.sym_tab=params[i]; i+=1
            opt.bar_detection=params[i]; i+=1
            opt.fill=params[i]; i+=1
            opt.specklesize=params[i]; i+=1
            opt.smooth=params[i]; i+=1
            opt.corr_rot=params[i]; i+=1
            opt.remove_cpy=params[i]; i+=1
            image=load_image(opt.imagefile)
            wx.BeginBusyCursor()

            if opt.corr_rot:
                print "rotation correction...",
                sys.stdout.flush()
                image=image.deskew()
                print "done"
            if image.data.pixel_type != ONEBIT:
                image = image.to_onebit()
            if opt.specklesize <1:
                print "despeckling...",
                sys.stdout.flush()
                opt.specklesize=7
                print "done"
                
            image.despeckle(opt.specklesize)
            image.invert()
            image.despeckle(opt.specklesize)
            image.invert()
            
            if opt.smooth < 3:
                print "smoothing...",
                sys.stdout.flush()
                image=image.smooth_onebit(0,opt.smooth)
                print "done"
            if opt.remove_cpy:
                print "removing copy borders...",
                sys.stdout.flush()
                image=image.remove_copy_border()
                print "done"

            wx.EndBusyCursor()


            # create a classifier and load the database
            classifier=knn.kNNInteractive([], [\
                    'aspect_ratio',\
                    'moments',\
                    'volume64regions',\
                    'nrows_feature'\
                    ], 0)
            if opt.datafile:
                classifier.from_xml_filename(opt.datafile)

            #delete lines:
            germanpage=GermanLutePage(image,opt.fill,opt.bar_detection)
            image=germanpage.image.to_onebit()

            #gui.main_win.shell.GetLocals().update(locals())
            gui.main_win.shell.GetLocals().update({'image': image,\
                                   'classifier': classifier})

            # split the image and open the interactive classifier window
            ccs=image.cc_analysis()
            
            if not opt.sym_tab:
                classifier.display(ccs, image, german_symbol_table)
            else:
                f=open(opt.sym_tab,"r")
                sym_tab=[]
                sym_file=f.readlines()
                for string in sym_file:
                    string=string.replace("\n","")
                    sym_tab.append(string)

                classifier.display(ccs, image, sym_tab)

#-------------------------------------------------------------------------------------

        def createGermanLutePageObj(self, event):
            #
            # let the user choose the parameters
            #
            clp_module="GermanLutePage"

            dialog=Args([FileOpen("Image file", "", "*.*"),\
                         Check("Heuristic bar detection",default=True),\
                         Check("Fill holes in lines"),\
                         Check("Write into log_file"),\
                         Int("Despecle size")],\
                         "Create a %s object" %( clp_module))
            params=dialog.show()

            if params != None:
                if params[0] != None:
                    #
                    # load the image here and load it
                    # into the gamera shell, too. this is
                    # done for checking whether
                    # it is a onebit image or not.
                    #
                    filename=params[0]
                    imagename=path.basename(params[0])
                    imagename=imagename.split('.')[0]
                    # substitute special characters
                    imagename=re.sub('[^a-zA-Z0-9]', '_',\
                            imagename)
                    imagename=re.sub('^[0-9]', '_',\
                            imagename)
                    image=load_image(filename)

                    self._shell.run('%s = load_image(r"%s")'\
                            % (imagename, filename))

                    ## if image.data.pixel_type != ONEBIT:
##                         self._shell.run("%s = %s.to_onebit()"\
##                                 % (imagename,\
##                                 imagename))
                    if image.data.pixel_type != ONEBIT:
                        image = image.to_onebit()
                    speckle_size=params[-1]
                    if speckle_size <3:
                        print "despeckling...",
                        sys.stdout.flush()
                        specklesize=7
                        print "done"

                    image.despeckle(specklesize)
                    image.invert()
                    image.despeckle(specklesize)
                    image.invert()

                    # still exists in the gamera shell
                    del image

                    # choose a name for the variable in
                    # the GUI
                    if clp_module.startswith("GermanLutePage"):
                        name=var_name.get("german",\
                            self._shell.locals)
                    else:
                        name=var_name.get("german",\
                            self._shell.locals)

                    if name is "" or name is None:
                        return

                    # create an instance of the specified
                    # CreateGermanLutePage class
                    self._shell.run("%s = %s.%s(%s, %d, %d, %d)"\
                            % (name,\
                            self.label,\
                            clp_module,\
                            imagename,\
                            params[1], params[2],params[3]))
                    # deletes image after handing over to constructor
                    self._shell.run("del %s"%(imagename))
                    
#--------------------------------------------------------------------------



    #--------------------------------------------------------------------------
    #---------------------------French Tab-------------------------------------
    #--------------------------------------------------------------------------

        def openfrenchtrainingsdialog(self,event):

            from gamera.toolkits.otr.otr_staff import TabPage
            from gamera.toolkits.otr.otr_glyph import symbol_table

            # start options dialog
            dialog=Args([FileOpen("Image file", "", "*.*"),
                         FileOpen("Training data file for reuse", "", "*.xml"),
                         Check("Smooth"),
                         Check("Deskew"),
                         Check("Remove pictures"),
                         Check("Despeckle"),
                         Check("Symbols on lines"),
                         Int("Lines per staff", default=0),
                         Check("Show original image")],
			name="Select training options")
            params=dialog.show()

            if not params or not params[0]:
                return

            # wxPython
            import wx

            # map parameters
            i=0
            imagefile=params[i]; i+=1
            trainingdata=params[i]; i+=1
            smooth=params[i]; i+=1
            deskew=params[i]; i+=1
            depict=params[i]; i+=1
            despec=params[i]; i+=1
            onlines=params[i]; i+=1
            numlines=params[i]; i+=1
            show_staves=params[i]; i+=1

            wx.BeginBusyCursor()
            image=load_image(imagefile)

            if deskew:
                    image=image.deskew()
            if smooth:
                    if image.data.pixel_type != GREYSCALE:
                            image = image.to_greyscale()
                    image=image.smoothen()
            if image.data.pixel_type != ONEBIT:
                    image = image.to_onebit()
            if depict:
                    image = image.remove_pictures()
            if despec:
                    size = image.most_frequent_run('black','vertical') * 2
                    image.despeckle(size)

            if onlines:
                    onlines=1
            else:
                    onlines=0

            # create a TabPage object and remove the staves
            tab=TabPage(image, online=onlines)
            tab.remove_staves(num_lines=numlines)

            wx.EndBusyCursor()

            # create a classifier and load the database
            classifier=knn.kNNInteractive([], [\
                            'aspect_ratio',\
                            'moments',\
                            'volume64regions',\
                            'nrows_feature'\
                            ], 0)
            if trainingdata:
                    classifier.from_xml_filename(trainingdata)

            #gui.main_win.shell.GetLocals().update(locals())
            gui.main_win.shell.GetLocals().update({'image': image,\
                                                   'classifier': classifier})

            # split the image and open the interactive classifier window
            ccs=tab.image.cc_analysis()
            if show_staves:
                    classifier.display(ccs, image, symbol_table)
            else:
                    classifier.display(ccs, tab.image, symbol_table)



    GermanLutePageModuleIcon.register()

