# -*- mode: python; indent-tabs-mode: nil; tab-width: 4 -*-
# vim: set tabstop=4 shiftwidth=4 expandtab:

##############################################################################
# german_glyph.py
#
# Christine Pranzas, 2007
#
##############################################################################

## german_page
## is a page for german lute tabulature
from gamera import toolkit
from math import sqrt
import random
import sys
from math import sqrt,log
from german_page import *
from to_abcmusic import *
import plugins
from gamera.core import *
from gamera.args import *
from gamera.gui import gui
from gamera import toolkit
from gamera.toolkits.otr.plugins import *
from gamera.core import *
from gamera.gui import has_gui
if has_gui.has_gui:
    from gamera.gui import var_name
    #import wx

german_symbol_table=["fret.c6.f0","fret.c6.f1", "fret.c6.f2","fret.c6.f3","fret.c6.f4","fret.c6.f5","fret.c6.f6","fret.c6.f7",
    "fret.c1.f0","fret.c1.f1", "fret.c1.f2","fret.c1.f3","fret.c1.f4","fret.c1.f5","fret.c1.f6","fret.c1.f7",
    "fret.c2.f0","fret.c2.f1", "fret.c2.f2","fret.c2.f3","fret.c2.f4","fret.c2.f5","fret.c2.f6","fret.c2.f7",
    "fret.c3.f0","fret.c3.f1", "fret.c3.f2","fret.c3.f3","fret.c3.f4","fret.c3.f5","fret.c3.f6","fret.c3.f7",
    "fret.c4.f0","fret.c4.f1", "fret.c4.f2","fret.c4.f3","fret.c4.f4","fret.c4.f5","fret.c4.f6","fret.c4.f7",
    "fret.c5.f0","fret.c5.f1", "fret.c5.f2","fret.c5.f3","fret.c5.f4","fret.c5.f5","fret.c5.f6","fret.c5.f7",
    "flag.1.1", "flag.2.1","flag.2.2", "flag.4.1","flag.4.2","flag.4.4",
    "flag.8.1", "flag.8.2","flag.8.4",
    "flag.16.1","flag.16.2","flag.16.4",
    "flag.32.1","flag.32.2","flag.32.4",
    "flag.64.1","flag.64.2","flag.64.4",
    "flag.128.1","flag.128.2","flag.128.4",
    "deco.repeat","above.stroke","bottom.stroke",
    "above.higher_fet","deco.rest.1",
    "deco.meter_C_3","deco.meter_C","right.deco.ternuto",
    "bar", "bar_part","dot", "flag.fermata", "trash"
    ]


class GermanGlyph:
##############################################################################
##############################################################################
    """Performs the semantic grouping of a classified list of German 
tablature glyphs and generates abc output for abctab2ps_.
All methods for semantic grouping are called in the constructor.
The resulting abc code is stored in the property *song*.

.. _abctab2ps: http://www.lautengesellschaft.de/cdmm/
"""

    ######################################################################
    # constructor
    # 
    def __init__(self, img,ccs,lines, bars,line_height,distance_between_two_lines,character_height,dots,title,debug ):
        """The constructor already does all the semantic analysis. Consequently
there is usually no need to call any other methods of *GermanGlyph*.

Signature:

  ``__init__(image,ccs,lines_mb,bars,line_height,distance_between_two_lines,character_height,dots,title,debug)``

with

  *image*:
    Onebit or greyscale image that is processed.

  *ccs*:
    List of classified glyphs to be semantically grouped.

  *line_height*, *distance_between_two_lines*, *charcter_height*:
    The corresponding properties of *GermanLutePage*.

  *line_mbs*:
    The least quare fit parameters (m,b) of all lines. This is the
    property *straight_lines* of *GermanLutePage*.

  *dots*:
    If dots shall be goruped to flags. This should be usually set to ``True``,
    unless dotted flags have been trained as groups.

  *title*:
    Song title in abc output.

  *debug*:
    Debug level (verbosity). Can be =,1, or 2.
"""

        if img.data.pixel_type!=ONEBIT:
            self.original_image=img.to_onebit()
        else:
            self.original_image=img.image_copy()

        if img.ul_x!=0 or img.ul_y!=0:
            img2=Image((0,0),img.size,ONEBIT)
            img2.or_image(self.image,True)
            self.original_image=img2

        self.image=self.original_image.image_copy()
        self.ccs=ccs
        self.lines=lines
        self.bars=bars
        self.dots=dots
        self.debug=debug
        self.title=title
        self.debug=debug
        self.number_of_accord_groups=0
        self.number_of_bars=0
        self.number_of_symbols=0
        self.number_of_lines=0
        self.line_height=line_height
        self.distance_between_two_lines=distance_between_two_lines
        self.character_height=character_height
        self.song=""
        self.page=[]
        self.page=self.get_sorted_page()
        self.song=self.generate_abc_output()


#-------------------------------------------------------------
    #get the value of a notegroup in a flag_group
    def get_flag_value(self,flag_member):
        count_flags=0
        if len(flag_member)%2==0:
            #last element is missing. therefore append one
            for i,flm in enumerate(flag_member):
                if i%2==0 and i!=0:
                    count_flags+=int(flag_member[i])
            count_flags+=1

        else:
            for i,flm in enumerate(flag_member):
                if i%2==0 and i!=0:
                    #print flag_member
                    count_flags+=int(flag_member[i])


        return count_flags

#-------------------------------------------------------------
    #sort all symbols into line_groups, bar_groups, flag_groups and note_groups.
    
    def get_sorted_page(self):

        #sort glyphs into line_groups
        self.lines.sort(lambda a,b:cmp(a[1],b[1]))

        #sort elements into line_groups, so that the symbols get sorted
        #------------------------------------------
        groups=[]
        sorted_accords=[]
        sorted_lines=[]
        left_right=[]
        dot=[]

        for line in range(len(self.lines)+1):
            groups.append([])
            sorted_accords.append([])
            sorted_lines.append([])
            left_right.append([])
            dot.append([])

        for cc in self.ccs:
            for nr in range(len(self.lines)):

                if cc.center_y<self.lines[0][0]*cc.center_x+self.lines[0][1]:
                    groups[0].append(cc)
                    break
                if cc.center_y>self.lines[-1][0]*cc.center_x+self.lines[-1][1]:
                    groups[len(self.lines)].append(cc)
                    break

                if cc.center_y>self.lines[nr-1][0]*cc.center_x+self.lines[nr-1][1] and cc.center_y<self.lines[nr][0]*cc.center_x+self.lines[nr][1]:
                    groups[nr].append(cc)
                    break


        #get characteristic values for the image (character_height and bar_height)
        for group in groups:
            group.sort(lambda a,b: cmp(a.ul_y,b.ul_y))

        bar_height=0
        c_height=0
        bar_black_pixel=0
        c_black_picel=0
        for group in groups:
            for cc in group:
                if "bar" in cc.get_main_id().split("."):
                    if bar_height==0:
                        bar_height=cc.height
                    else:
                        bar_height=(bar_height+cc.height)/2
                elif "fret" in cc.get_main_id().split("."):
                    if c_height==0:
                        c_height=cc.height
                    else:
                        c_height=(c_height+cc.height)/2

        if self.debug>0:
            print "c_height,bar_height", c_height,bar_height

        #check if there are bar_parts detected:
        # put bars to one group if they overlapp in one line_group
        gr_number=0
        while gr_number<len(groups):
            idx=0
            bar_groups=[]
            while idx<len(groups[gr_number]):
                cc=groups[gr_number][idx]
                if "bar" in cc.get_main_id().split(".") or "bar_part" in cc.get_main_id().split("."):

                    if bar_groups==[]:
                        bar_groups.append([cc])
                    else:
                        bar_idx=0
                        while bar_idx<len(bar_groups):
                            left= max(bar_groups[bar_idx][-1].ul_x,cc.ul_x)
                            right=min(bar_groups[bar_idx][-1].lr_x,cc.lr_x)
                            ov1=right-left
                            dist=abs(bar_groups[bar_idx][-1].center_x-cc.center_x)

                            top= max(bar_groups[bar_idx][0].ul_y,cc.ul_y)
                            bottom=min(bar_groups[bar_idx][-1].lr_y,cc.lr_y)
                            ov2=abs(bottom-top)
                            if ov2>self.line_height*2 and dist<c_height:
                                if bar_groups[bar_idx][-1].height<cc.height:
                                    del bar_groups[bar_idx][-1]
                                    bar_groups[bar_idx].append(cc)
                                    break
                            elif dist<c_height or ov1>0:
                                bar_groups[bar_idx].append(cc)
                                break
                            bar_idx+=1
                        else:
                            bar_groups.append([cc])
                    del groups[gr_number][idx]
                else:
                    idx+=1

            gr_idx=0
            #classify new the new bars
            while gr_idx< len(bar_groups):
                bar=union_images(bar_groups[gr_idx])
                bar.classify_heuristic("bar")
                if bar.height>c_height :
                    groups[gr_number].append(bar)
                gr_idx+=1        
            gr_number+=1



        missing_line=False

        for group in groups:
            group.sort(lambda a,b: cmp(a.center_x,b.center_x))

        #search if line one line is missing
        #therefore find all flags and check if some flags are not in the same row
        #search all flags belonging to the same row
        # if there are more rows than one and if in the other row groups are more
        # than 3 Flags a new line is added
        #---------------------------------------------------------
        for group in groups:
            flag_array=[]
            for cc in group:
                if "flag" in cc.get_main_id():
                    if flag_array==[]:
                        flag_array.append([cc])
                    else:
                        flag_idx=0
                        while flag_idx<len(flag_array):

                            top=max(cc.ul_y,flag_array[flag_idx][-1].ul_y)
                            bottom=min(cc.lr_y,flag_array[flag_idx][-1].lr_y)
                            overlapp_y=bottom-top

                            left=max(cc.ul_x,flag_array[flag_idx][-1].ul_x)
                            right=min(cc.lr_x,flag_array[flag_idx][-1].lr_x)
                            overlapp_x=right-left

                            dist=abs(cc.center_y-flag_array[flag_idx][-1].center_y)

                            if overlapp_y>0 or dist<self.distance_between_two_lines/2:
                                flag_array[flag_idx].append(cc)
                                break
                            flag_idx+=1
                        else:
                            flag_array.append([cc])

            
            if len(flag_array)>0:
                big_flag_idx = 0
                y_short=self.image.height
                flag_idx=0
                while flag_idx<len(flag_array):
                    if len(flag_array[flag_idx])<4:
                        del flag_array[flag_idx]
                    elif flag_array[flag_idx][0].center_y<y_short:
                        big_flag_idx=flag_idx
                        y_short=flag_array[flag_idx][0].center_y
                        flag_idx+=1
                    else:
                        flag_idx+=1
                        
                if len(flag_array)>0:
                    del flag_array[big_flag_idx]
            if len(flag_array)>0:
                flag_idx=0
               
                while flag_idx<len(flag_array):
                    y_high=self.image.height
                    if len(flag_array[flag_idx])>4:
                        
                        sec_flag_idx=0
                        while sec_flag_idx<len(flag_array[flag_idx]):
                            if flag_array[flag_idx][sec_flag_idx].ul_y<y_high:
                                y_high=flag_array[flag_idx][sec_flag_idx].ul_y
                            sec_flag_idx+=1
                        #add missing line with median slope and the highest pixel of the flag_group
                        if self.debug>0:
                            print " Add a new line at y= ",y_high
                        median_slope=-1
                        for line in self.lines:
                            if median_slope==-1:
                                median_slope=line[0]
                            else:
                                median_slope=(median_slope+line[0])/2
                        if median_slope==-1:
                            median_slope=0
                        line=[median_slope,y_high]
                        self.lines.append(line)
                        self.lines.sort(lambda a,b:cmp(a[1],b[1]))
                        self.bars.insert(self.lines.index(line),[])
                        missing_line=True

                    flag_idx+=1


        # if line is missing group the symbols new into groups
        #---------------------------------------------------------------------
        if missing_line==True:

            groups=[]
            sorted_accords=[]
            sorted_lines=[]
            left_right=[]
            dot=[]

            for line in range(len(self.lines)+1):
                groups.append([])
                sorted_accords.append([])
                sorted_lines.append([])
                left_right.append([])
                dot.append([])

            for cc in self.ccs:
                for nr in range(len(self.lines)):
                    if cc.center_y<self.lines[0][0]*cc.center_x+self.lines[0][1]:
                        groups[0].append(cc)
                        break
                    if cc.center_y>self.lines[-1][0]*cc.center_x+self.lines[-1][1]:
                        groups[len(self.lines)].append(cc)
                        break

                    if cc.center_y>self.lines[nr-1][0]*cc.center_x+self.lines[nr-1][1] and cc.center_y<self.lines[nr][0]*cc.center_x+self.lines[nr][1]:
                        groups[nr].append(cc)
                        break             


        #sort elements into their bar groups. Sort Symbols in line_groups so that the symbol with the smallest x-value is first.
        #begin to sort all elements in the first bar group until a bar is found. A new group is created and than the next symbols
        #will be sorted into this group...
        #do this for each line_group
        #when a flag is found, sort it in the right position into the array all symbols will be grouped in
        # when special signs are found they are grouped into special arrays. Those signs will later processed.  
        #---------------------------------------------------

        for n1 in range(len(groups)):
            groups[n1].sort(lambda a,b:cmp(a.ul_x,b.ul_x))
            sorted_lines[n1].append([])
            sorted_accords[n1].append([])
            left_right[n1].append([])
            dot[n1].append([])
            
            sorted_lines[n1].append([])
            sorted_accords[n1].append([])
            left_right[n1].append([])
            dot[n1].append([])
            
        if self.bars!=[]:
            
            for bar in self.bars:
                bar.sort(lambda a,b: cmp(a[0].center_x,b[0].center_x))
        
            gr_number=0
            while gr_number<len(sorted_lines):
                bar_groups=0
                bar_idx=0
                flag=False
                while flag==False:
                    flag=True
                    cc_number=0
                    flag_idx=0
                     
                    while cc_number<len( groups[gr_number]):
                        if "bar" in groups[gr_number][cc_number].get_main_id().split("."):
                            #open new bar-group
                            del groups[gr_number][cc_number]
                            sorted_lines[gr_number].append([])
                            sorted_accords[gr_number].append([])
                            left_right[gr_number].append([])
                            dot[gr_number].append([])
                            
                            flag=False
                            break
                        elif self.bars!=[] and bar_idx<len(self.bars[gr_number]) and len(self.bars[gr_number][bar_idx])>0 and groups[gr_number][cc_number].center_x>self.bars[gr_number][bar_idx][len(self.bars[gr_number][bar_idx])/2].center_x:
                           
                            sorted_lines[gr_number].append([])
                            sorted_accords[gr_number].append([])
                            left_right[gr_number].append([])
                            dot[gr_number].append([])
                            flag=False
                            bar_idx+=1
                            break
                        elif "flag" in groups[gr_number][cc_number].get_main_id().split("."):
                            sorted_accords[gr_number][bar_groups].append([[groups[gr_number][cc_number]]])
                            flag=False
                            #open new bar-group  
                            del groups[gr_number][cc_number]
                        elif "left" in groups[gr_number][cc_number].get_main_id().split(".") or "right" in groups[gr_number][cc_number].get_main_id().split("."):
                            left_right[gr_number][bar_groups].append(groups[gr_number][cc_number])
                            flag=False
                            del groups[gr_number][cc_number]
                        elif "dot" in groups[gr_number][cc_number].get_main_id().split("."):
                            flag=False
                            dot[gr_number][bar_groups].append(groups[gr_number][cc_number])
                            del groups[gr_number][cc_number]
                        else:

                            sorted_lines[gr_number][bar_groups].append(groups[gr_number][cc_number])
                            del groups[gr_number][cc_number]

                    bar_groups+=1
                gr_number+=1
        else:
            gr_number=0
            while gr_number<len(sorted_lines):
                bar_groups=0
                bar_idx=0
                while bar_groups<len(sorted_lines[gr_number]):
                    cc_number=0
                    flag_idx=0
                    while cc_number<len( groups[gr_number]):

                        if "bar" in groups[gr_number][cc_number].get_main_id().split("."):
                            #open new bar-group
                            del groups[gr_number][cc_number]
                            break
                        elif "flag" in groups[gr_number][cc_number].get_main_id().split("."):
                            sorted_accords[gr_number][bar_groups].append([[groups[gr_number][cc_number]]])
                            #open new bar-group  
                            del groups[gr_number][cc_number]
                        elif "left" in groups[gr_number][cc_number].get_main_id().split(".") or "right" in groups[gr_number][cc_number].get_main_id().split("."):
                            left_right[gr_number][bar_groups].append(groups[gr_number][cc_number])
                            del groups[gr_number][cc_number]
                        elif "dot" in groups[gr_number][cc_number].get_main_id().split("."):
                            dot[gr_number][bar_groups].append(groups[gr_number][cc_number])
                            del groups[gr_number][cc_number]
                        else:

                            sorted_lines[gr_number][bar_groups].append(groups[gr_number][cc_number])
                            del groups[gr_number][cc_number]

                    bar_groups+=1
                gr_number+=1
            


        #------------------- sort all groups into extra note_groups------------
        #----------------------------------------------------------------------
        # try to group all notes. A note group consists of notes who overlapp each other vertically.
        # For each Symbol try to find that symbol with which it overlapps most. A graph with nodes and
        # egdes is arranged. After that add all note_groups together. If two symbols in one group overlapp
        # an error is occured. So that an other algorithem must be used. Therefore find the row with most
        #symbols and divide the symbols in rows and columns. For each element not belonging to the row with
        # the most symbols, a column in the row is acquiered. Only one element from a row belongs to a column.  

        for gr_number,sorted_line in enumerate(sorted_lines):
            for bar_number,groupccs  in enumerate(sorted_line):
            #links=kanten; numbers=knoten; aufbauen eines ungerichteten graphen
                links=[]
                for i1,cc1 in enumerate(groupccs):
                    maxoverlap=0
                    maxoverlap_i=-1
                    dist=-1
                    for i2 in range(i1+1,len(groupccs)):
                        cc2=groupccs[i2]
                        left=max(cc1.ul_x,cc2.ul_x)
                        right=min(cc1.lr_x,cc2.lr_x)
                        ov=right-left

                        #save y-dist
                        top=max(cc1.ul_y,cc2.ul_y)
                        bottom=min(cc1.lr_y,cc2.lr_y)
                        ov2=abs(bottom-top)

                        if ov>maxoverlap and (ov>0.1*cc1.width or ov>0.1*cc2.width):
                            maxoverlap=ov
                            dist=ov2
                            maxoverlap_i=i2

                    if maxoverlap_i!=-1:
                        links.append([i1,maxoverlap_i,float(maxoverlap)/min((cc1.width,cc2.width))])
                # find for each node the achievable neighbornodes
                aqs=[]
                numbers=range(len(groupccs))

                #sort glyhs into their accord
                while len(numbers)>0:

                    n=numbers[0]
                    del numbers[0]
                    temp_links=[]
                    temp_numbers=[n]
                    #  breadth-first search
                    aq=[]
                    aq_part=[n]
                    while len(aq_part)>0:
                        link_n=0
                        while link_n<len(links):
                            if links[link_n][0]==aq_part[0]:
                                if links[link_n][1] in numbers:
                                    temp_links.append(links[link_n])
                                    temp_numbers.append(links[link_n][1])
                                    aq_part.append(links[link_n][1])
                                    numbers.remove(links[link_n][1])
                                    del links[link_n]
                                else:
                                    link_n+=1
                            elif links[link_n][1]==aq_part[0]:
                                if links[link_n][0] in numbers:
                                    temp_links.append(links[link_n])
                                    temp_numbers.append(links[link_n][0])
                                    aq_part.append(links[link_n][0])
                                    numbers.remove(links[link_n][0])
                                    del links[link_n]
                                else:
                                    link_n+=1
                            else:
                                link_n+=1

                        aq.append(groupccs[aq_part[0]])
                        del aq_part[0]

                    # Elements in one note_group are not supposed to overlapp. If they overlapp there is an error. The Elements are subdivided into their y-coordinates.
                    # new groups are formed
                    #--------------------------------------------------------------------------------------------------------------------

                    overlap_exists=False
                    fist_element=-1
                    sec_element=-1
                    for i1,num1 in enumerate(temp_numbers):
                        cc1=groupccs[num1]
                        if "dot" in cc1.get_main_id():
                            continue
                        for i2 in range(i1+1,len(temp_numbers)):

                            num2=temp_numbers[i2]
                            cc2=groupccs[num2]
                            if "dot" in cc1.get_main_id():
                                continue
                            overlap=min((cc1.lr_y,cc2.lr_y))-max((cc1.ul_y,cc2.ul_y))
                            if overlap>self.line_height:
                                first_element=num1
                                sec_element=num2

                                overlap_exists=True
                    splited_group=[]

                    split_gr=[]

                    if overlap_exists:
                        aq.sort(lambda a,b:cmp(a.center_y,b.center_y))
                        # subdivide elements in x-direction. So that they are arranged in panes.
                        # aufteilung in x richtung scheiben 
                        for a in aq:
                            if splited_group==[]:
                                first_element=a
                                splited_group.append([a])
                                continue


                            top=max(a.ul_y,first_element.ul_y)
                            bottom=min(a.lr_y,first_element.lr_y)
                            ov=bottom-top
                            if ov>self.line_height:
                                splited_group[-1].append(a)
                            else:
                                first_element=a
                                splited_group.append([a])

                       
                        for split in splited_group:
                            split.sort(lambda a,b:cmp(a.center_x,b.center_x))
                        # they group with the most elements, is jused to the the basisgroup
                        longest=-1
                        nr_longest=-1
                        for nr,split in enumerate(splited_group):
                            if len(split)>longest:
                                longest=len(split)
                                nr_longest=nr
                        # get the x-values of the basisgroup
                        center_array=[]
                        for cc in splited_group[nr_longest]:
                            center_array.append(cc.center_x)


                        for i in range(longest):
                            split_gr.append([])

                        #sort each element to its group. Only one element on the same pane gets a place in the notegroup.
                        for split1 in splited_group:
                            append_number=[]
                            for cc in split1:
                                most_fit=self.image.width
                                most_fit_number=-1
                                next_fit_number=-1
                                for number,center in enumerate(center_array):
                                    if number in append_number and "dot" not in cc.get_main_id():
                                        continue
                                    if abs(cc.center_x-center)<most_fit and abs(cc.center_x-center)<self.character_height*2:
                                        next_fit=most_fit_number
                                        most_fit=abs(cc.center_x-center)
                                        most_fit_number=number

                                if most_fit_number!=-1:
                                    split_gr[most_fit_number].append(cc)
                                    if "dot" not in cc.get_main_id():
                                        append_number.append(most_fit_number)
                                else:
                                    print "Error."
                                    #self.image.highlight(cc,RGBPixel(255,0,0))


                    if overlap_exists==False:
                        aqs.append(aq)
                    else:
                        for split in split_gr:
                            aqs.append(split)


                #--------------------------------------------------------------------
                #---------- sort left_right into their groups -----------------------
                # the left_right symbols belong to the symbol on the right / left which is not a special sign
                # when such an element is found, the element gets a new attribute, which comprised the name of
                # the left_right symbol
                
                if left_right[gr_number][bar_number]!=[]:
                    for cclr in left_right[gr_number][bar_number]:

                        if "right" in cclr.get_main_id():

                            gr_idx=0
                            idx=-1
                            c_idx=-1
                            dist=self.image.width
                            while gr_idx<len(aqs):
                                sec_idx=0
                                while sec_idx< len(aqs[gr_idx]):
                                    sec_cclr=aqs[gr_idx][sec_idx]
                                    top=max(cclr.ul_y,sec_cclr.ul_y)
                                    bottom=min(cclr.lr_y,sec_cclr.lr_y)
                                    ov=bottom-top
                                    tdist=sec_cclr.center_x-cclr.center_x           
                                    if tdist>0 and dist>tdist and ov>0.5*cclr.width :
                                        idx=gr_idx
                                        c_idx=sec_idx
                                        dist=tdist
                                        break
                                    sec_idx+=1
                                gr_idx+=1
                            if idx!=-1:
                                if hasattr(aqs[idx][c_idx],"deco"):

                                    aqs[idx][c_idx].deco.append(cclr.get_main_id())
                                else:
                                    aqs[idx][c_idx].deco=[]
                                    aqs[idx][c_idx].deco.append(cclr.get_main_id())
                                aqs[idx].append(cclr)

                        elif "left" in cclr.get_main_id():
                            gr_idx=0
                            idx=-1
                            c_idx=-1
                            dist=self.image.width
                            while gr_idx<len(aqs):
                                sec_idx=0
                                while sec_idx< len(aqs[gr_idx]):
                                    sec_cclr=aqs[gr_idx][sec_idx]
                                    top=max(cclr.ul_y,sec_cclr.ul_y)
                                    bottom=min(cclr.lr_y,sec_cclr.lr_y)
                                    ov=bottom-top
                                    tdist=cclr.center_x- sec_cclr.center_x           
                                    if tdist>0 and dist>tdist and ov>0.5*cclr.width:
                                        idx=gr_idx
                                        c_idx=sec_idx
                                        dist=tdist
                                        break
                                    sec_idx+=1
                                gr_idx+=1
                            if idx!=-1:

                                if hasattr(aqs[idx][c_idx],"deco"):
                                    aqs[idx][c_idx].deco.append(cclr.get_main_id())
                                else:
                                    aqs[idx][c_idx].deco=[]
                                    aqs[idx][c_idx].deco.append(cclr.get_main_id())
                                aqs[idx].append(cclr)


                #---------------------------------------------------------
                # sort dots into their groups
                # dots may belong to a flag, so that the value of a flag is extended  about the half of its original value
                # or dots belong to note_groups. There they have no speacial meaning
                
                if dot[gr_number][bar_number]!=[] and self.dots==True:
                    #look if dot is member of a flag

                    d_idx=0
                    while d_idx<len(dot[gr_number][bar_number]):
                        d=dot[gr_number][bar_number][d_idx]
                        flag_idx=-1
                        save_dist=self.image.width
                        for fx,flag in enumerate(sorted_accords[gr_number][bar_number]):
                            top=max(d.ul_y,flag[0][0].ul_y)
                            bottom=min(d.ll_y,flag[0][0].ll_y)
                            overlapped=bottom-top
                            dist=d.center_x-flag[0][0].ul_x
                            if overlapped>=d.height and dist>0 and dist<self.character_height and dist<save_dist  and ("flag" in flag[0][0].get_main_id().split(".") or "fermata" in flag[0][0].get_main_id().split(".")):
                                save_dist=dist
                                flag_idx=fx
                        if flag_idx!=-1:
                            sorted_accords[gr_number][bar_number][flag_idx][0][0].dot=True
                            dot[gr_number][bar_number].remove(dot[gr_number][bar_number][d_idx])
                            d_idx-=1
                        d_idx+=1


                    #sort dots to their notegroups
                    
                    for d in dot[gr_number][bar_number]:
                            gr_idx=0
                            idx=-1
                            dist=self.image.width
                            most_ov=-1
                            ov_idx=-1
                            while gr_idx<len(aqs):
                                sec_idx=0
                                while sec_idx< len(aqs[gr_idx]):
                                    sec_d=aqs[gr_idx][sec_idx]

                                    left=max(d.ul_x,sec_d.ul_x)
                                    right=min(d.lr_x,sec_d.lr_x)
                                    overlapp=right-left
                                    tdist=abs(sec_d.center_x-d.center_x)           
                                    if overlapp>0 and most_ov<overlapp:
                                        ov_idx=gr_idx
                                        most_ov=overlapp
                                        ov=True

                                    elif  dist>tdist and tdist<self.character_height:
                                        idx=gr_idx
                                        dist=tdist
                                        break
                                    sec_idx+=1
                                gr_idx+=1
                            if ov_idx!=-1:
                                aqs[ov_idx].append(d)
                            elif idx!=-1:
                                aqs[idx].append(d)

                #---------------
                #group above and bottom to their symbols
                # when a stroke belongs to a number, that means that this element belongs on the 6th chord on an other fret
                # when a higher_fret belongs to a character it means that this note is suposed to be  five frets higher than
                # the original note
                
                idx=0
                while idx<len(aqs):
                    flag_gr=aqs[idx]
                    cnt=0
                    while (cnt < len(flag_gr)):
                        if "higher_fet" in flag_gr[cnt].get_main_id() or "stroke" in flag_gr[cnt].get_main_id():
                            sec_cnt=0
                            nearest=-1
                            nr=-1
                            while(sec_cnt <len(flag_gr)):
                                if cnt==sec_cnt:
                                    sec_cnt+=1
                                    continue
                                dist1=flag_gr[sec_cnt].ul_y-flag_gr[cnt].ll_y
                                dist2=flag_gr[cnt].ll_y-flag_gr[sec_cnt].ul_y
                                dist1x=flag_gr[sec_cnt].center_x-flag_gr[cnt].center_x
                                dist2x=flag_gr[cnt].center_x-flag_gr[sec_cnt].center_x
                                if "above" in flag_gr[cnt].get_main_id() and flag_gr[cnt].center_y <flag_gr[sec_cnt].center_y and (nearest==-1 or (dist1<nearest and dist1>=0)) and "above" not in flag_gr[sec_cnt].get_main_id() and "right" not in flag_gr[sec_cnt].get_main_id() and "left" not in flag_gr[sec_cnt].get_main_id() and "above" not in flag_gr[sec_cnt].get_main_id() and "bottom" not in flag_gr[sec_cnt].get_main_id():
                                    nearest=dist1
                                    nr=sec_cnt
                                    
                                elif "bottom" in flag_gr[cnt].get_main_id() and flag_gr[sec_cnt].center_y< flag_gr[cnt].center_y and (nearest==-1 or (dist2<nearest and dist2>0)) and "bottom" not in flag_gr[sec_cnt].get_main_id() and "right" not in flag_gr[sec_cnt].get_main_id() and "left" not in flag_gr[sec_cnt].get_main_id() and "above" not in flag_gr[sec_cnt].get_main_id() and "bottom" not in flag_gr[sec_cnt].get_main_id():
                                    nearest=dist2
                                    nr=sec_cnt
                                sec_cnt+=1

                            if nearest!=-1:
                                if ".c6." in flag_gr[nr].get_main_id():
                                    new_cc=union_images([flag_gr[nr],flag_gr[cnt]])
                                    #new classification
                                    new_cc.classify_heuristic(flag_gr[cnt].get_main_id())
                                    if nr<cnt:
                                        flag_gr.remove(flag_gr[cnt])
                                        flag_gr.remove(flag_gr[nr])
                                    else:
                                        flag_gr.remove(flag_gr[nr])
                                        flag_gr.remove(flag_gr[cnt])
                                    flag_gr.append(new_cc)
                                    cnt=0
                                    idx=0

                                elif ".f0" in flag_gr[nr].get_main_id():
                                    new_cc=union_images([flag_gr[nr],flag_gr[cnt]])
                                    name=""
                                    if ".c" in flag_gr[nr].get_main_id() and ".f" in flag_gr[nr].get_main_id():
                                        fl=flag_gr[nr].get_main_id().split(".")
                                        c=fl[1][-1]
                                        if c=="5":
                                            name="fret.c6.f0"
                                        elif c=="4":
                                            name="fret.c6.f1"
                                        elif c=="3":
                                            name="fret.c6.f2"
                                        elif c=="2":
                                            name="fret.c6.f3"
                                        elif c=="1":
                                            name="fret.c6.f4"
                                        else:
                                            name=flag_gr[nr].get_main_id()
                                    else:
                                        
                                        name=flag_gr[nr].get_main_id()

                                    #new classification
                                    new_cc.classify_heuristic(name)
                                    if nr<cnt:
                                        flag_gr.remove(flag_gr[cnt])
                                        flag_gr.remove(flag_gr[nr])
                                    else:
                                        flag_gr.remove(flag_gr[nr])
                                        flag_gr.remove(flag_gr[cnt])

                                    flag_gr.append(new_cc)
                                    cnt=0
                                    idx=0
                                elif "fret" in flag_gr[nr].get_main_id():
                                    #new classification
                                    new_cc=union_images([flag_gr[nr],flag_gr[cnt]])
                                    name=flag_gr[nr].get_main_id().split(".")
                                    f=int(name[-1][-1])+5
                                    name[-1]="f%d"%(f)
                                    fret_name=".".join(name)
                                    new_cc.classify_heuristic(fret_name)
                                    if nr<cnt:
                                        flag_gr.remove(flag_gr[cnt])
                                        flag_gr.remove(flag_gr[nr])
                                    else:
                                        flag_gr.remove(flag_gr[nr])
                                        flag_gr.remove(flag_gr[cnt])
                                    flag_gr.append(new_cc)
                                    cnt=0
                                    idx=0
                        
                        elif "bottom" in flag_gr[cnt].get_main_id() or "above" in flag_gr[cnt].get_main_id():
                            sec_cnt=0
                            nearest=-1
                            nr=-1
                            while(sec_cnt <len(flag_gr)):
                                if cnt==sec_cnt:
                                    sec_cnt+=1
                                    continue
                                dist1=flag_gr[sec_cnt].ul_y-flag_gr[cnt].ll_y
                                dist2=flag_gr[cnt].ll_y-flag_gr[sec_cnt].ul_y
                                dist1x=flag_gr[sec_cnt].center_x-flag_gr[cnt].center_x
                                dist2x=flag_gr[cnt].center_x-flag_gr[sec_cnt].center_x
                                if "above" in flag_gr[cnt].get_main_id() and flag_gr[cnt].center_y <flag_gr[sec_cnt].center_y and (nearest==-1 or (dist1<nearest and dist1>=0)) and "above" not in flag_gr[sec_cnt].get_main_id() and "right" not in flag_gr[sec_cnt].get_main_id() and "left" not in flag_gr[sec_cnt].get_main_id() and "above" not in flag_gr[sec_cnt].get_main_id() and "bottom" not in flag_gr[sec_cnt].get_main_id():
                                    nearest=dist1
                                    nr=sec_cnt
                                    
                                elif "bottom" in flag_gr[cnt].get_main_id() and flag_gr[sec_cnt].center_y< flag_gr[cnt].center_y and (nearest==-1 or (dist2<nearest and dist2>0)) and "bottom" not in flag_gr[sec_cnt].get_main_id() and "right" not in flag_gr[sec_cnt].get_main_id() and "left" not in flag_gr[sec_cnt].get_main_id() and "above" not in flag_gr[sec_cnt].get_main_id() and "bottom" not in flag_gr[sec_cnt].get_main_id():
                                    nearest=dist2
                                    nr=sec_cnt
                                sec_cnt+=1

                            if nearest!=-1:
                                if hasattr(flag_gr[nr],"deco"):

                                    flag_gr[nr].deco.append(flag_gr[cnt].get_main_id())
                                else:
                                    flag_gr[nr].deco=[]
                                    flag_gr[nr].deco.append(flag_gr[cnt].get_main_id())
                                

                                



                             
                        cnt+=1
                    idx+=1

                #----------------



                #sort each note_group to its flag
                #when a note_group does not overlapp with a flag or the flag_group is allready full,
                #search on the left and on the right side if there is a empty place for the note_group
                #if there is none, the group is intersperese addicted to the x-coordinate to the right
                #position in the array
                #----------------------------
                #---------------------------
                #-------------------------
                not_assigned=[]
                for aqi in aqs:
                    left=aqi[0].ul_x
                    right=aqi[0].lr_x
                    has_rest_repeat=False
                    for cc in aqi:
                        if "left" in cc.get_main_id() or "right" in cc.get_main_id() or "dot" in cc.get_main_id().split("."):
                            continue
                        if "rest" in cc.get_main_id().split(".") or  "repeat" in cc.get_main_id().split("."):
                            has_rest_repeat=True
                        if cc.ul_x<left:
                            left=cc.ul_x
                        if cc.lr_x>right:
                            right=cc.lr_x
                    sort_idx=0
                    most_ov=-5
                    most_idx=-1
                    #dist1=self.image.width
                    while sort_idx<len(sorted_accords[gr_number][bar_number]) and has_rest_repeat!=True:
                        left2=max(sorted_accords[gr_number][bar_number][sort_idx][0][0].ul_x,left)
                        right2=min(sorted_accords[gr_number][bar_number][sort_idx][0][0].lr_x,right)
                        overlapp=right2-left2
                      
                        not_full=False

                        if "flag" in sorted_accords[gr_number][bar_number][sort_idx][0][0].get_main_id().split("."):
                            count_flags=0
                            if not "fermata" in sorted_accords[gr_number][bar_number][sort_idx][0][0].get_main_id().split(".") :

                                flag_member=sorted_accords[gr_number][bar_number][sort_idx][0][0].get_main_id().split(".")

                                count_flags=self.get_flag_value(flag_member)


                            if "fermata" in sorted_accords[gr_number][bar_number][sort_idx][0][0].get_main_id().split(".") :
                                not_full=True
                            elif len(sorted_accords[gr_number][bar_number][sort_idx])<count_flags+1:
                                    not_full=True

                        if (overlapp>most_ov)  and (not_full or  "flag" not in sorted_accords[gr_number][bar_number][sort_idx][0][0].get_main_id()):
                            most_ov=overlapp
                            most_idx=sort_idx


                        sort_idx+=1

                    if most_idx!=-1:
                        sorted_accords[gr_number][bar_number][most_idx].append(aqi)
                    elif has_rest_repeat==True:
                        not_assigned.append(aqi)
                    else:
                        not_assigned.append(aqi)

                # try to find a flag for not assigned groups
                #-------------------------------------------
                
                if not_assigned!=[]:
                    na_idx=0
                    while na_idx<len(not_assigned):
                        na=not_assigned[na_idx]

                        left=na[0].ul_x
                        right=na[0].lr_x
                        for cc in na:
                            if "left" in cc.get_main_id() or "right" in cc.get_main_id() or "dot" in cc.get_main_id().split("."):
                                continue
                            if cc.ul_x<left:
                                left=cc.ul_x
                            if cc.lr_x>right:
                                right=cc.lr_x

                        center=(left+right)/2
                        #find flag on left and right side

                        sort_idx=0
                        #dist1=self.image.width
                        while sort_idx<len(sorted_accords[gr_number][bar_number])-1:
                            count_flags=-2

                            if "flag" in sorted_accords[gr_number][bar_number][sort_idx][0][0].get_main_id():

                                if not "fermata" in sorted_accords[gr_number][bar_number][sort_idx][0][0].get_main_id().split(".") :

                                    flag_member=sorted_accords[gr_number][bar_number][sort_idx][0][0].get_main_id().split(".")
                                    count_flags=self.get_flag_value(flag_member)

                                    if center<sorted_accords[gr_number][bar_number][0][0][0].center_x:

                                        #accord before first flag
                                        if "flag" in sorted_accords[gr_number][bar_number][0][0][0].get_main_id():
                                            flag_member=sorted_accords[gr_number][bar_number][0][0][0].get_main_id().split(".")
                                            count_flags=self.get_flag_value(flag_member)
                                        if len(sorted_accords[gr_number][bar_number][0])<count_flags+1 and abs(right-sorted_accords[gr_number][bar_number][0][0][0].ul_x)<self.character_height*2:
                                            #print "############## accord first flag",na[0].center_x,na[0].center_y,len(sorted_accords[gr_number][bar_number][0]),count_flags+1
                                            sorted_accords[gr_number][bar_number][0].insert(0,na)
                                            not_assigned.remove(not_assigned[na_idx])
                                            na_idx-=1
                                        break
                                    elif center>sorted_accords[gr_number][bar_number][-1][0][0].center_x:

                                        #accord behind last flag
                                        if "flag" in sorted_accords[gr_number][bar_number][-1][0][0].get_main_id():
                                            flag_member=sorted_accords[gr_number][bar_number][-1][0][0].get_main_id().split(".")
                                            count_flags=self.get_flag_value(flag_member)
                                        if len(sorted_accords[gr_number][bar_number][-1])<count_flags+1 and abs(left-sorted_accords[gr_number][bar_number][-1][0][0].ur_x)<self.character_height*2:
                                            #print "----- accord behind flag",na[0].center_x,na[0].center_y
                                            sorted_accords[gr_number][bar_number][-1].append(na)
                                            not_assigned.remove(not_assigned[na_idx])
                                            na_idx-=1
                                        break
                                    elif center>sorted_accords[gr_number][bar_number][sort_idx][0][0].center_x and center<sorted_accords[gr_number][bar_number][sort_idx+1][0][0].center_x:


                                        #when accord is inbetween two flags
                                        if len(sorted_accords[gr_number][bar_number][sort_idx])<count_flags+1 and abs(left-sorted_accords[gr_number][bar_number][sort_idx][0][0].ur_x)<self.character_height*2:
                                            #belongs to left flag
                                            #print "$$$$$$$$$$$$$belongs to left_flag",na[0].center_x,na[0].center_y
                                            sorted_accords[gr_number][bar_number][sort_idx].append(na)
                                            not_assigned.remove(not_assigned[na_idx])
                                            na_idx-=1
                                            break

                                        if "flag" in sorted_accords[gr_number][bar_number][sort_idx+1][0][0].get_main_id():
                                            count_flags=0
                                            if not "fermata" in sorted_accords[gr_number][bar_number][sort_idx][0][0].get_main_id().split(".") :


                                                count_flags=self.get_flag_value(sorted_accords[gr_number][bar_number][sort_idx+1][0][0].get_main_id().split("."))

                                                
                                                if len(sorted_accords[gr_number][bar_number][sort_idx+1])<count_flags+1 and abs(right-sorted_accords[gr_number][bar_number][sort_idx+1][0][0].ul_x)<self.character_height:
                                                    sorted_accords[gr_number][bar_number][sort_idx+1].insert(1,na)
                                                    not_assigned.remove(not_assigned[na_idx])
                                                    na_idx-=1
                                                    #print "############# belongs to right flag",na[0].center_x,na[0].center_y, na[0].get_main_id()
                                                    break


                            sort_idx+=1
                        na_idx+=1

                # try to find a place  where notegroup fits
                #-------------------------------------------
                if not_assigned!=[]:

                    na_idx=0
                    while na_idx<len(not_assigned):
                        na=not_assigned[na_idx]
                        fits=False
                        f_idx=0
                        while (f_idx<len(sorted_accords[gr_number][bar_number])) and fits==False:
                            note_idx=0
                            while note_idx<len(sorted_accords[gr_number][bar_number][f_idx]):

                                if "flag" in sorted_accords[gr_number][bar_number][f_idx][0][0].get_main_id():
                                    symbol=sorted_accords[gr_number][bar_number][f_idx][note_idx][0]
                                    if "flag" in sorted_accords[gr_number][bar_number][f_idx][note_idx][0].get_main_id():
                                        pass
                                    elif na[0].center_x< symbol.center_x and note_idx==1:
                                        sorted_accords[gr_number][bar_number].insert(f_idx,[na])
                                        fits=True
                                        break
                                    elif na[0].center_x< symbol.center_x :
                                        sorted_accords[gr_number][bar_number][f_idx].insert(note_idx,na)
                                        fits=True
                                        break
                                        
                                else:
                                    symbol=sorted_accords[gr_number][bar_number][f_idx][note_idx][0]
                                    if na[0].center_x< symbol.center_x:
                                        sorted_accords[gr_number][bar_number].insert(f_idx,[na])
                                        fits=True
                                        break

                                                                                                                                

                                note_idx+=1
                            if fits==True:
                                break
                            f_idx+=1
                        else:
                            sorted_accords[gr_number][bar_number].append([na])
                        na_idx+=1



        groups=0
        symbols=0
        bars=0
        lines=0

        for gr_number,sorted_accord in enumerate(sorted_accords):
            if len(sorted_accords[gr_number])>0:
                    lines+=1
            for bar_number,accord in enumerate(sorted_accord):
                if len(sorted_accords[gr_number][bar_number])>0:
                    bars+=1
                for acc in accord:
                    groups+=1
                    color=RGBPixel(random.randint(34,224),random.randint(34,224),random.randint(34,224))
                    for flag in acc:

                        for note in flag:
                            symbols+=1
                           # color=RGBPixel(random.randint(34,224),random.randint(34,224),random.randint(34,224))
                           # self.image.highlight(note,color)
                    #self.image.highlight(acc[0][0],RGBPixel(255,255,0))
        self.number_of_accord_groups=groups
        self.number_of_bars=bars
        self.number_of_symbols=symbols
        self.number_of_lines=lines
        
        #delete all empty bar_groups at the end
        gr_number=0
        while gr_number<len(sorted_accords):
            bar_number=len(sorted_accords[gr_number])-1
            while bar_number>=0:
                if len(sorted_accords[gr_number][bar_number])==0:
                    del sorted_accords[gr_number][bar_number]
                    self.number_of_bars-=1
                    bar_number-=1
                else:
                    break
            gr_number+=1

        return sorted_accords


    #---------------------------------------------------
    def mark_flag_groups(self):
        """Returns an RGB image with all symbols belonging to the same flag
highlighted in the same color. This is useful for debugging.
"""
        #mark all symbols belonging to the same flag in one color.
        mark_flag_groups=self.image.to_rgb()
        for gr_number,sorted_accord in enumerate(self.page):
            for bar_number,accord in enumerate(sorted_accord):
                for acc in accord:
                    color=RGBPixel(random.randint(34,224),random.randint(34,224),random.randint(34,224))
                    for flag in acc:
                        for note in flag:
                            mark_flag_groups.highlight(note,color)
        return mark_flag_groups

    #---------------------------------------------------
    def mark_note_groups(self):
        """Returns an RGB image with different note groups highlighted
in different colors. This is useful for debugging.
"""
        #mark all note_groups in diffenent colors
        mark_note_groups=self.image.to_rgb()

        for gr_number,sorted_accord in enumerate(self.page):
            flag_color=RGBPixel(random.randint(34,224),random.randint(34,224),random.randint(34,224))
            for bar_number,accord in enumerate(sorted_accord):
                for acc in accord:
                    for flag in acc:
                        color=RGBPixel(random.randint(34,224),random.randint(34,224),random.randint(34,224))
                        for note in flag:
                            if "flag" in note.get_main_id():
                                mark_note_groups.highlight(note,RGBPixel(0,255,0))
                            else:
                                mark_note_groups.highlight(note,color)
        return mark_note_groups
        




    #---------------------------------------------------
    def generate_abc_output(self):
        #generate the abc output and return it as a string

        song=""
        song= song+"%\n"+\
               "%%tabaddflags    1\n"+\
               "%%tabrhstyle     grid\n"+\
               "%%taballflags    yes\n"+\
               "%%tabbrummer     123\n"+\
               "%%titleleft      yes\n"+\
               "%%staffsep       30.0pt\n"+\
               "%%tabfontgerman  deFraktur\n"+\
               "%%musicspace     0.50cm\n"+\
               "%\n"

        tabtype="germantab"
        div=2
        title=self.title
        n_items=self.number_of_symbols
        n_chords=self.number_of_bars
        
        # add header fields
        song =song+ "X:1\n"+\
               "T:" + title + "\n"+\
               "C:" + str(n_items) + " symbols in " + str(n_chords) + " chords.\n"+\
               "L:1/"+str(div)+"\n"+\
               "K:" + str(tabtype) + "\n"



        def get_value(flag,gr_number):
            #get the rhythm value of a notegroup
            if "fermata" in flag.get_main_id():
                return "!fermata!"
            if "_part" in flag.get_main_id():
                return ""
            # get rhythm value for a note_group
            name=flag.get_main_id().split(".")[1:]
            
            if len(name)%2==1:
                name.append(1)
            i=0
            s=[]
            #generate list of rhythm values and take that one which belongs to the note_group
            while i<len(name):
                s.extend([int(name[i])]*int(name[i+1]))
                i+=2
            if gr_number>=len(s):
                return ""
            if hasattr(flag,"dot") and gr_number==0:
                #print "------------------------ dot belongs to flag"
                return "3/"+str(s[gr_number]*2)
            elif int(log(s[gr_number],2))==0:
                return "1"
            else:
                return "/"*int(log(s[gr_number],2))
        
        def get_accord_string(note_group,first_in_bar_group,value,value_old):
            #get the accord string for each note_group, with value
            #fret0=a, fret1=b,fret2=c....
            # when a comma is before a note for example two commas, that the note belongs the the thrid chord. The first and the second chord will not be played.
            #gernate an array of commas
            string=[","]*6
            stars=[]
            #walk through the note_group and replace all commas with the aquiered notes
            for cc in note_group:
                name=cc.get_main_id()
                if "c" in name and ".f" in name and not "_part" in name and not "_group" in name:
                    c=name[name.index("c")+1:name.index(".f")]
                    f=name[name.index(".f")+2:]
                    string[int(c)-1]="abcdefghijklmnopqrstuvwxyz"[int(f)]
                    if hasattr(cc,"deco"):
                        for star in cc.deco:
                            if "ternuto" in star:
                                string[int(c)-1]="*"+string[int(c)-1]

            string="".join(string)
            #delete last element out of the array until a comma is found
            while len(string)>0 and string[-1]==",":
                string=string[:-1]
            if string=="":
                return ""
            # the rhythm value must be wrote, when it is the first symbol of a group or when the rhythm value has changed
            if (first_in_bar_group or value!=value_old) and value!="!fermata!":
                string+=value
            #when more than one note belongs to a note_group, the notes must combine in brackets
            if len(note_group)>1:
                string="["+string+"]"
            if value=="!fermata!":
                string="!fermata!"+string
            return string
        
        value=""
       
        for line in self.page:
            song=song+"%\n"
            for bar_number,bar_group in enumerate(line):
                for flag_group in bar_group:
                    if "flag" in flag_group[0][0].get_main_id() or "fermata" in flag_group[0][0].get_main_id():
                        if len(flag_group)==1:
                            if "fermata" in flag_group[0][0].get_main_id().split("."):
                                song+="!fermata!"
                                continue
                            flag_name=flag_group[0][0].get_main_id().split(".")
                            if int(log(float(flag_name[1]),2))==0:
                                string="z1"
                            else:
                                string="z"+"/"*int(log(float(flag_name[1]),2))
                            
                            song+=string
                        else:
                            flag=flag_group[0]
                            for gr_number,note_group in enumerate(flag_group[1:]):
                                value_old=value
                                value=get_value(flag[0],gr_number)
                                s=get_accord_string(note_group,gr_number==0,value,value_old)
                                song=song+s
                    elif "deco.rest" in flag_group[0][0].get_main_id():
                        rest=flag_group[0][0].get_main_id().split(".")
                        if int(log(float(rest[-1]),2))==0:
                            string="z1"
                        else:
                            string="z"+"/"*int(log(float(rest[-1]),2))
                        song=song+string
                        
                    elif "deco.repeat" in flag_group[0][0].get_main_id():
                        song+="::"
                    elif ".meter" in flag_group[0][0].get_main_id():
                        if "3" in  flag_group[0][0].get_main_id():
                            song=song+"[M:3]"
                        else:
                            song=song+"[M:C]"
                    else:
                        for gr_number,note_group in enumerate(flag_group):
                            value_old=value
                            s=get_accord_string(note_group,gr_number==0,value_old,value_old)
                            song=song+s
                        
                    song=song+" "
                    
                
                if bar_number==len(line)-1:
                    song=song+"|\n"
                    #pass
                else:
                    song=song+"|\\\n\t"
        
        return song


    #--------------------------------------------------------------------------------

    def to_abcmusic(self, tuning="gdAFCG,", key='auto', factor=0):
        """Gives an output of the sorted list ``self.glyphs`` as an 
*abc music*-formatted string.

Signature:

  ``to_abcmusic(tuning=\"gdAFCG,\", key=\"auto\", factor=0)``

with

  *tuning*:
    The tuning of each course, from top to bottom. Each note is specified
    as *abc* code (sharp = prefix '^', flat = prefix '_', octave up = postfix
    single quote, octave down = postfix comma)

  *key*:
    Key signature, e.g. the value ``Eb`` stands for *E flat Major* or ``Cm``
    for *C Minor*. When set to ``auto`` the key is guessed automatically.

  *factor*:
    Logarithmic length factor for slowing down. For example, when set to
    ``2`` the sixteenth notes are printed as quarter notes. A negative
    integer leads to a speedup.

The returned string can be given as input to the external program
`abctab2ps`__.

.. __: http://www.lautengesellschaft.de/cdmm/

"""
        abcmusic=ToAbcmusic(self.page,"german",tuning,key,factor)
        song=abcmusic.abc_music

        return song

