#!/usr/bin/python

#
# otr_prepare - preprocessing for optical tablature recognition
#               this file is part of the OTR toolkit for gamera
#
# Authors:  Christoph Dalitz  2004-2013
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 


# initialization
import sys
import os
import re

from gamera.core import *
from gamera.toolkits.otr.plugins import *

#
# parse command line arguments
#
class c_opt:
	infile = ""
	outfile = ""
	depict = 0
        nlines = 0
	smooth = 0
	deskew = 0
	despec = 0
	suffix = ""
	def error_exit(self):
		usage = "Usage:\n" + \
				"\t" + sys.argv[0] + " -o <outfile> [options] <infile>\n" + \
				"Options:\n" + \
				"\t-smooth  perform smoothing\n" + \
				"\t-deskew  perform deskewing\n" + \
				"\t-depict  remove pictures\n" + \
				"\t-n <num> number of staff lines (for -depict)\n" + \
				"\t-despec  remove CC's smaller than staffline_height * 2\n" + \
				"<outfile> must end with .png, .tif or .tiff\n" + \
				"this extension determines the output image format\n"
		sys.stderr.write(usage)
		sys.exit(1)
opt = c_opt()

i = 1
while i < len(sys.argv):
	if sys.argv[i] == "-o":
		i += 1; opt.outfile = sys.argv[i]
	elif sys.argv[i] == '-depict':
		opt.depict = 1
	elif sys.argv[i] == "-n":
		i += 1; opt.nlines = int(sys.argv[i])
	elif sys.argv[i] == '-deskew':
		opt.deskew = 1
	elif sys.argv[i] == '-smooth':
		opt.smooth = 1
	elif sys.argv[i] == '-despec':
		opt.despec = 1
	elif sys.argv[i][0] == '-':
		opt.error_exit()
	else:
		opt.infile = sys.argv[i]
	i += 1

# some plausi checks
if (opt.infile == ""):
	sys.stderr.write("No infile given.\n")
	opt.error_exit()
if not os.path.isfile(opt.infile):
	sys.stderr.write("File '" + opt.infile + "' not found.\n")
	opt.error_exit()
if (opt.outfile == ""):
	sys.stderr.write("No outfile given.\n")
	opt.error_exit()
m = re.search("\.[^.]+$", opt.outfile)
if m:
	opt.suffix = m.group(0)
else:
	opt.suffix = ""
if (opt.suffix != ".tiff") and (opt.suffix != ".tif") \
   and (opt.suffix != ".png"):
	sys.stderr.write("Unknown outfile extension '" + opt.suffix + "\n")
	opt.error_exit()

# prevent gamera from parsing command line
sys.argv = []

init_gamera()

# do quested preprocessing
image=load_image(opt.infile)

if opt.deskew:
	if image.data.pixel_type != GREYSCALE:
		image = image.to_greyscale()
	print "Deskew..."
	image = image.deskew()

if opt.smooth:
	if image.data.pixel_type != GREYSCALE:
		image = image.to_greyscale()
	print "Smoothen..."
	image = image.smoothen()

if image.data.pixel_type != ONEBIT:
	image = image.to_onebit()

if opt.depict:
	if opt.nlines == 0:
            factor = 5
        else:
            factor = opt.nlines
	print "Remove pictures (height_factor = %d)..." % factor
	image = image.remove_pictures(factor)

if opt.despec:
	size = image.most_frequent_run('black', 'vertical') * 2
	print "Despeckle (size = %d)..." % size
	image.despeckle(size)

# save to requested format
if opt.suffix == ".tiff" or opt.suffix == ".tif":
	image.save_tiff(opt.outfile)
else:
	image.save_PNG(opt.outfile)
